// Licensed to the Apache Software Foundation (ASF) under one
// or more contributor license agreements.  See the NOTICE file
// distributed with this work for additional information
// regarding copyright ownership.  The ASF licenses this file
// to you under the Apache License, Version 2.0 (the
// "License"); you may not use this file except in compliance
// with the License.  You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

// Code generated by the FlatBuffers compiler. DO NOT EDIT.

package flatbuf

import "strconv"

type SparseMatrixCompressedAxis int16

const (
	SparseMatrixCompressedAxisRow    SparseMatrixCompressedAxis = 0
	SparseMatrixCompressedAxisColumn SparseMatrixCompressedAxis = 1
)

var EnumNamesSparseMatrixCompressedAxis = map[SparseMatrixCompressedAxis]string{
	SparseMatrixCompressedAxisRow:    "Row",
	SparseMatrixCompressedAxisColumn: "Column",
}

var EnumValuesSparseMatrixCompressedAxis = map[string]SparseMatrixCompressedAxis{
	"Row":    SparseMatrixCompressedAxisRow,
	"Column": SparseMatrixCompressedAxisColumn,
}

func (v SparseMatrixCompressedAxis) String() string {
	if s, ok := EnumNamesSparseMatrixCompressedAxis[v]; ok {
		return s
	}
	return "SparseMatrixCompressedAxis(" + strconv.FormatInt(int64(v), 10) + ")"
}
