// <auto-generated>
//  automatically generated by the FlatBuffers compiler, do not modify
// </auto-generated>

namespace Apache.Arrow.Flatbuf
{

internal enum Precision : short
{
 HALF = 0,
 SINGLE = 1,
 DOUBLE = 2,
};


}
