// <auto-generated>
//  automatically generated by the FlatBuffers compiler, do not modify
// </auto-generated>

namespace Apache.Arrow.Flatbuf
{

/// ----------------------------------------------------------------------
/// Endianness of the platform producing the data
internal enum Endianness : short
{
 Little = 0,
 Big = 1,
};


}
