// <auto-generated>
//  automatically generated by the FlatBuffers compiler, do not modify
// </auto-generated>

namespace Apache.Arrow.Flatbuf
{

internal enum TimeUnit : short
{
 SECOND = 0,
 MILLISECOND = 1,
 MICROSECOND = 2,
 NANOSECOND = 3,
};


}
