/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.arrow.flight.auth2;

import org.apache.arrow.flight.CallHeaders;
import org.apache.arrow.flight.grpc.CredentialCallOption;

/**
 * Interface for client side header parsing and conversion to CredentialCallOption.
 */
public interface ClientHeaderHandler {
  /**
   * Parses the incoming headers and converts them into a CredentialCallOption.
   * @param incomingHeaders Incoming headers to parse.
   * @return An instance of CredentialCallOption.
   */
  CredentialCallOption getCredentialCallOptionFromIncomingHeaders(CallHeaders incomingHeaders);
  
  /**
   * An client header handler that does nothing.
   */
  ClientHeaderHandler NO_OP = new ClientHeaderHandler() {
    @Override
    public CredentialCallOption getCredentialCallOptionFromIncomingHeaders(CallHeaders incomingHeaders) {
      return null;
    }
  };
}
