# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

class RollingWindowTest < Test::Unit::TestCase
  include Helper::Fixture

  def setup
    raw_table = {
      :number => Arrow::Int32Array.new([1, -2, nil, 4, 6, 3]),
    }
    @table = Arrow::Table.new(raw_table)
  end

  test("#lag") do
    assert_equal(<<-ARRAY.chomp, @table.window.lag(:number).to_s)
[
  null,
  -3,
  null,
  null,
  2,
  -3
]
    ARRAY
  end
end
