// Licensed to the Apache Software Foundation (ASF) under one
// or more contributor license agreements.  See the NOTICE file
// distributed with this work for additional information
// regarding copyright ownership.  The ASF licenses this file
// to you under the Apache License, Version 2.0 (the
// "License"); you may not use this file except in compliance
// with the License.  You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing,
// software distributed under the License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
// KIND, either express or implied.  See the License for the
// specific language governing permissions and limitations
// under the License.

namespace plasma.flatbuf;

// Object information data structure.
table ObjectInfo {
  // Object ID of this object.
  object_id: string;
  // Number of bytes the content of this object occupies in memory.
  data_size: long;
  // Number of bytes the metadata of this object occupies in memory.
  metadata_size: long;
  // Number of clients using the objects.
  ref_count: int;
  // Unix epoch of when this object was created.
  create_time: long;
  // How long creation of this object took.
  construct_duration: long;
  // Hash of the object content. If the object is not sealed yet this is
  // an empty string.
  digest: string;
  // Specifies if this object was deleted or added.
  is_deletion: bool;
}
