# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

if_version <- function(version, op = `==`) {
  op(packageVersion("arrow"), version)
}

if_version_less_than <- function(version) {
  if_version(version, op = `<`)
}

skip_if_version_less_than <- function(version, msg) {
  if (if_version(version, `<`)) {
    skip(msg)
  }
}

skip_if_version_equals <- function(version, msg) {
  if (if_version(version, `==`)) {
    skip(msg)
  }
}
