// Licensed to the Apache Software Foundation (ASF) under one
// or more contributor license agreements.  See the NOTICE file
// distributed with this work for additional information
// regarding copyright ownership.  The ASF licenses this file
// to you under the Apache License, Version 2.0 (the
// "License"); you may not use this file except in compliance
// with the License.  You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing,
// software distributed under the License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
// KIND, either express or implied.  See the License for the
// specific language governing permissions and limitations
// under the License.

#ifndef CREATEPROCESS_MANIFEST_RESOURCE_ID
#define CREATEPROCESS_MANIFEST_RESOURCE_ID  1
#endif
#ifndef RT_MANIFEST
#define RT_MANIFEST  24
#endif

CREATEPROCESS_MANIFEST_RESOURCE_ID RT_MANIFEST
BEGIN
  "<?xml version=""1.0"" encoding=""UTF-8"" standalone=""yes""?>"
  "<assembly xmlns=""urn:schemas-microsoft-com:asm.v1"" manifestVersion=""1.0"">"
    "<assemblyIdentity type=""win32"" name=""ArrowUtilityTest"" version=""1.1.1.1""/>"
    "<application xmlns=""urn:schemas-microsoft-com:asm.v3"">"
      "<windowsSettings xmlns:ws2=""http://schemas.microsoft.com/SMI/2016/WindowsSettings"">"
        "<ws2:longPathAware>true</ws2:longPathAware>"
      "</windowsSettings>"
    "</application>"
    "<trustInfo xmlns=""urn:schemas-microsoft-com:asm.v3"">"
      "<security>"
        "<requestedPrivileges>"
          "<requestedExecutionLevel level=""asInvoker"" uiAccess=""false""/>"
        "</requestedPrivileges>"
      "</security>"
    "</trustInfo>"
  "</assembly>"
END

