# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

module Plasma
  class Client
    alias_method :initialize_raw, :initialize
    private :initialize_raw
    def initialize(socket_path, options=nil)
      socket_path = socket_path.to_path if socket_path.respond_to?(:to_path)
      if options
        options_raw = options
        options = ClientOptions.new
        options_raw.each do |key, value|
          setter = "#{key}="
          options.__send__(setter, value) if options.respond_to?(setter)
        end
      end
      initialize_raw(socket_path, options)
    end
  end
end
