// <auto-generated>
//  automatically generated by the FlatBuffers compiler, do not modify
// </auto-generated>

namespace Apache.Arrow.Flatbuf
{

internal enum CompressionType : sbyte
{
 LZ4_FRAME = 0,
 ZSTD = 1,
};


}
