// <auto-generated>
//  automatically generated by the FlatBuffers compiler, do not modify
// </auto-generated>

namespace Apache.Arrow.Flatbuf
{

internal enum IntervalUnit : short
{
 YEAR_MONTH = 0,
 DAY_TIME = 1,
};


}
