/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.arrow.flight;

import org.apache.arrow.flight.impl.Flight;

import com.google.protobuf.ByteString;

/**
 * An opaque object that can be used to filter a list of streams available from a server.
 *
 * <p>This is a POJO wrapper around the protobuf Criteria message.
 */
public class Criteria {

  public static Criteria ALL = new Criteria((byte[]) null);

  private final byte[] bytes;

  public Criteria(byte[] bytes) {
    this.bytes = bytes;
  }

  Criteria(Flight.Criteria criteria) {
    this.bytes = criteria.getExpression().toByteArray();
  }

  /**
   * Get the contained filter criteria.
   */
  public byte[] getExpression() {
    return bytes;
  }

  Flight.Criteria asCriteria() {
    Flight.Criteria.Builder b = Flight.Criteria.newBuilder();
    if (bytes != null) {
      b.setExpression(ByteString.copyFrom(bytes));
    }

    return b.build();
  }
}
