/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.arrow.flight;

/**
 * An exception raised from a Flight RPC.
 *
 * <p>In service implementations, raising an instance of this exception will provide clients with a more detailed
 * message and error code.
 */
public class FlightRuntimeException extends RuntimeException {
  private final CallStatus status;

  /**
   * Create a new exception from the given status.
   */
  FlightRuntimeException(CallStatus status) {
    super(status.description(), status.cause());
    this.status = status;
  }

  public CallStatus status() {
    return status;
  }

  @Override
  public String toString() {
    String s = getClass().getName();
    return String.format("%s: %s: %s", s, status.code(), status.description());
  }
}
