// Licensed to the Apache Software Foundation (ASF) under one or more
// contributor license agreements. See the NOTICE file distributed with
// this work for additional information regarding copyright ownership.
// The ASF licenses this file to You under the Apache License, Version 2.0
// (the "License"); you may not use this file except in compliance with
// the License.  You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

using System.Collections.Generic;
using System.Linq;

namespace Apache.Arrow.Types
{
    public enum UnionMode
    {
        Sparse,
        Dense
    }

    public sealed class UnionType : ArrowType
    {
        public override ArrowTypeId TypeId => ArrowTypeId.Union;
        public override string Name => "union";

        public UnionMode Mode { get; }
        
        public IEnumerable<byte> TypeCodes { get; }

        public UnionType(
            IEnumerable<Field> fields, IEnumerable<byte> typeCodes,
            UnionMode mode = UnionMode.Sparse)
        {
            TypeCodes = typeCodes.ToList();
            Mode = mode;
        }

        public override void Accept(IArrowTypeVisitor visitor) => Accept(this, visitor);
    }
}
