// Licensed to the Apache Software Foundation (ASF) under one
// or more contributor license agreements.  See the NOTICE file
// distributed with this work for additional information
// regarding copyright ownership.  The ASF licenses this file
// to you under the Apache License, Version 2.0 (the
// "License"); you may not use this file except in compliance
// with the License.  You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package arrow

type BinaryType struct{}

func (t *BinaryType) ID() Type       { return BINARY }
func (t *BinaryType) Name() string   { return "binary" }
func (t *BinaryType) String() string { return "binary" }
func (t *BinaryType) binary()        {}

type StringType struct{}

func (t *StringType) ID() Type       { return STRING }
func (t *StringType) Name() string   { return "utf8" }
func (t *StringType) String() string { return "utf8" }
func (t *StringType) binary()        {}

var (
	BinaryTypes = struct {
		Binary BinaryDataType
		String BinaryDataType
	}{
		Binary: &BinaryType{},
		String: &StringType{},
	}
)
