# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

class ListDataTypeTest < Test::Unit::TestCase
  sub_test_case(".new") do
    test("Arrow::Field") do
      field = Arrow::Field.new(:tag, :string)
      assert_equal("list<tag: string>",
                   Arrow::ListDataType.new(field).to_s)
    end

    test("name: String") do
      assert_equal("list<tag: string>",
                   Arrow::ListDataType.new(name: "tag", type: :string).to_s)
    end

    test("field: Arrow::Field") do
      field = Arrow::Field.new(:tag, :string)
      assert_equal("list<tag: string>",
                   Arrow::ListDataType.new(field: field).to_s)
    end

    test("field: Hash") do
      field_description = {name: "tag", type: :string}
      assert_equal("list<tag: string>",
                   Arrow::ListDataType.new(field: field_description).to_s)
    end

    test("Arrow::DataType") do
      data_type = Arrow::BooleanDataType.new
      assert_equal("list<item: bool>",
                   Arrow::ListDataType.new(data_type).to_s)
    end

    test("String") do
      assert_equal("list<item: bool>",
                   Arrow::ListDataType.new("boolean").to_s)
    end

    test("Symbol") do
      assert_equal("list<item: bool>",
                   Arrow::ListDataType.new(:boolean).to_s)
    end

    test("[data type name, additional information]") do
      assert_equal("list<item: time32[ms]>",
                   Arrow::ListDataType.new([:time32, :milli]).to_s)
    end

    test("type: Symbol") do
      assert_equal("list<item: bool>",
                   Arrow::ListDataType.new(type: :boolean).to_s)
    end
  end
end
