// Licensed to the Apache Software Foundation (ASF) under one
// or more contributor license agreements.  See the NOTICE file
// distributed with this work for additional information
// regarding copyright ownership.  The ASF licenses this file
// to you under the Apache License, Version 2.0 (the
// "License"); you may not use this file except in compliance
// with the License.  You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing,
// software distributed under the License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
// KIND, either express or implied.  See the License for the
// specific language governing permissions and limitations
// under the License.

import { isTypedArray } from 'apache-arrow';

describe('isTypedArray', () => {
    test('works for typed arrays', () => {
        expect(isTypedArray(new Int8Array())).toBeTruthy();
        expect(isTypedArray(new Int32Array())).toBeTruthy();
        expect(isTypedArray(new BigInt64Array())).toBeTruthy();
    });

    test('does not recognize arrays, buffers, or data views', () => {
        expect(isTypedArray(new Array([1, 2, 3]))).toBeFalsy();
        expect(isTypedArray(new ArrayBuffer(10))).toBeFalsy();
        expect(isTypedArray(new DataView(new ArrayBuffer(10)))).toBeFalsy();
    });
});
