// <auto-generated>
//  automatically generated by the FlatBuffers compiler, do not modify
// </auto-generated>

namespace Apache.Arrow.Flatbuf
{

/// ----------------------------------------------------------------------
/// Top-level Type value, enabling extensible type-specific metadata. We can
/// add new logical types to Type without breaking backwards compatibility
internal enum Type : byte
{
 NONE = 0,
 Null = 1,
 Int = 2,
 FloatingPoint = 3,
 Binary = 4,
 Utf8 = 5,
 Bool = 6,
 Decimal = 7,
 Date = 8,
 Time = 9,
 Timestamp = 10,
 Interval = 11,
 List = 12,
 Struct_ = 13,
 Union = 14,
 FixedSizeBinary = 15,
 FixedSizeList = 16,
 Map = 17,
 Duration = 18,
 LargeBinary = 19,
 LargeUtf8 = 20,
 LargeList = 21,
};


}
