{#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#}
Total number of JIRA tickets assigned to version {{ release.version }}: {{ release.issues|length }}

Total number of applied patches since version {{ release.previous.version }}: {{ release.commits|length }}

Patches with assigned issue in version {{ release.version }}:
{% for issue, commit in within -%}
 - {{ commit.url }} {{ commit.title }}
{% endfor %}

Patches with assigned issue outside of version {{ release.version }}:
{% for issue, commit in outside -%}
 - {{ commit.url }} {{ commit.title }}
{% endfor %}

Patches in version {{ release.version }} without a linked issue:
{% for commit in nojira -%}
 - {{ commit.url }} {{ commit.title }}
{% endfor %}

JIRA issues in version {{ release.version }} without a linked patch:
{% for issue in nopatch -%}
 - https://issues.apache.org/jira/browse/{{ issue.key }}
{% endfor %}
