# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

module Gandiva
  class ExpressionBuilder
    class Record
      def initialize(schema)
        @schema = schema
      end

      def respond_to_missing?(name, include_private)
        return true if @schema[name]
        super
      end

      def method_missing(name, *args)
        return super unless args.empty?
        self[name] || super
      end

      def [](name)
        field = @schema[name]
        if field
          Field.new(field)
        else
          nil
        end
      end
    end
  end
end
