# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

class DictionaryArrayTest < Test::Unit::TestCase
  sub_test_case("instance methods") do
    def setup
      @values = ["a", "b", "c", "b", "a"]
      @string_array = Arrow::StringArray.new(@values)
      @array = @string_array.dictionary_encode
    end

    test("#[]") do
      assert_equal(@values, @array.to_a)
    end

    test("#get_value") do
      assert_equal([
                     @values[0],
                     @values[3],
                   ],
                   [
                     @array.get_value(0),
                     @array.get_value(3),
                   ])
    end
  end
end
