// Licensed to the Apache Software Foundation (ASF) under one or more
// contributor license agreements. See the NOTICE file distributed with
// this work for additional information regarding copyright ownership.
// The ASF licenses this file to You under the Apache License, Version 2.0
// (the "License"); you may not use this file except in compliance with
// the License.  You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

using System;
using System.Buffers;
using System.Runtime.CompilerServices;
using System.Threading.Tasks;

namespace Apache.Arrow
{
    internal static class ArrayPoolExtensions
    {
        [MethodImpl(MethodImplOptions.AggressiveInlining)]
        public static void RentReturn(this ArrayPool<byte> pool, int length, Action<Memory<byte>> action)
        {
            byte[] array = null;

            try
            {
                array = pool.Rent(length);
                action(array.AsMemory(0, length));
            }
            finally
            {
                if (array != null)
                {
                    pool.Return(array);
                }
            }
        }

        [MethodImpl(MethodImplOptions.AggressiveInlining)]
        public static ValueTask RentReturnAsync(this ArrayPool<byte> pool, int length, Func<Memory<byte>, ValueTask> action)
        {
            byte[] array = null;

            try
            {
                array = pool.Rent(length);
                return action(array.AsMemory(0, length));
            }
            finally
            {
                if (array != null)
                {
                    pool.Return(array);
                }
            }
        }
    }
}
