# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

class SortOptionsTest < Test::Unit::TestCase
  sub_test_case("#initialize") do
    test("none") do
      options = Arrow::SortOptions.new
      assert_equal([],
                   options.sort_keys.collect(&:to_s))
    end

    test("-String, Symbol") do
      options = Arrow::SortOptions.new("-count", :age)
      assert_equal(["-count", "+age"],
                   options.sort_keys.collect(&:to_s))
    end
  end

  sub_test_case("instance methods") do
    setup do
      @options = Arrow::SortOptions.new
    end

    sub_test_case("#add_sort_key") do
      test("-String") do
        @options.add_sort_key("-count")
        assert_equal(["-count"],
                     @options.sort_keys.collect(&:to_s))
      end

      test("-String, Symbol") do
        @options.add_sort_key("-count", :desc)
        assert_equal(["--count"],
                     @options.sort_keys.collect(&:to_s))
      end

      test("SortKey") do
        @options.add_sort_key(Arrow::SortKey.new("-count"))
        assert_equal(["-count"],
                     @options.sort_keys.collect(&:to_s))
      end
    end
  end
end
