// Licensed to the Apache Software Foundation (ASF) under one
// or more contributor license agreements.  See the NOTICE file
// distributed with this work for additional information
// regarding copyright ownership.  The ASF licenses this file
// to you under the Apache License, Version 2.0 (the
// "License"); you may not use this file except in compliance
// with the License.  You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing,
// software distributed under the License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
// KIND, either express or implied.  See the License for the
// specific language governing permissions and limitations
// under the License.

include "Schema.fbs";

namespace org.apache.arrow.flatbuf;

/// ----------------------------------------------------------------------
/// Arrow File metadata
///

table Footer {
  version: org.apache.arrow.flatbuf.MetadataVersion;

  schema: org.apache.arrow.flatbuf.Schema;

  dictionaries: [ Block ];

  recordBatches: [ Block ];

  /// User-defined metadata
  custom_metadata: [ KeyValue ];
}

struct Block {

  /// Index to the start of the RecordBlock (note this is past the Message header)
  offset: long;

  /// Length of the metadata
  metaDataLength: int;

  /// Length of the data (this is aligned so there can be a gap between this and
  /// the metadata).
  bodyLength: long;
}

root_type Footer;
