/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.arrow.adapter.orc;

import java.util.Arrays;
import java.util.List;

/**
 * Wrapper for record batch meta and native memory.
 */
class OrcRecordBatch {
  final int length;

  /**
   * Nodes correspond to the pre-ordered flattened logical schema.
   */
  final List<OrcFieldNode> nodes;

  final List<OrcMemoryJniWrapper> buffers;

  /**
   * Construct a new instance.
   * @param length number of records included in current batch
   * @param nodes meta data for each fields
   * @param buffers buffers for underlying data
   */
  OrcRecordBatch(int length, OrcFieldNode[] nodes, OrcMemoryJniWrapper[] buffers) {
    this.length = length;
    this.nodes = Arrays.asList(nodes);
    this.buffers = Arrays.asList(buffers);
  }
}
