/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.arrow.vector.util;

import java.nio.charset.StandardCharsets;
import java.util.concurrent.TimeUnit;

import org.apache.arrow.memory.BufferAllocator;
import org.apache.arrow.memory.RootAllocator;
import org.apache.arrow.vector.IntVector;
import org.apache.arrow.vector.VarCharVector;
import org.openjdk.jmh.annotations.Benchmark;
import org.openjdk.jmh.annotations.BenchmarkMode;
import org.openjdk.jmh.annotations.Mode;
import org.openjdk.jmh.annotations.OutputTimeUnit;
import org.openjdk.jmh.annotations.Scope;
import org.openjdk.jmh.annotations.Setup;
import org.openjdk.jmh.annotations.State;
import org.openjdk.jmh.annotations.TearDown;
import org.openjdk.jmh.runner.Runner;
import org.openjdk.jmh.runner.RunnerException;
import org.openjdk.jmh.runner.options.Options;
import org.openjdk.jmh.runner.options.OptionsBuilder;

/**
 * Benchmarks for {@link TransferPair}.
 */
@State(Scope.Benchmark)
public class TransferPairBenchmarks {

  private static final int VECTOR_LENGTH = 1024;

  private static final int ALLOCATOR_CAPACITY = 1024 * 1024;

  private BufferAllocator allocator;

  private IntVector intVector;

  private VarCharVector varCharVector;

  /**
   * Setup benchmarks.
   */
  @Setup
  public void prepare() {
    allocator = new RootAllocator(ALLOCATOR_CAPACITY);
    intVector = new IntVector("intVector", allocator);
    varCharVector = new VarCharVector("varcharVector", allocator);

    intVector.allocateNew(VECTOR_LENGTH);
    varCharVector.allocateNew(VECTOR_LENGTH);

    for (int i = 0; i < VECTOR_LENGTH; i++) {
      if (i % 3 == 0) {
        intVector.setNull(i);
        varCharVector.setNull(i);
      } else {
        intVector.setSafe(i, i * i);
        varCharVector.setSafe(i, ("teststring" + i).getBytes(StandardCharsets.UTF_8));
      }
    }
    intVector.setValueCount(VECTOR_LENGTH);
    varCharVector.setValueCount(VECTOR_LENGTH);
  }

  /**
   * Tear down benchmarks.
   */
  @TearDown
  public void tearDown() {
    intVector.close();
    varCharVector.close();;
    allocator.close();
  }

  @Benchmark
  @BenchmarkMode(Mode.AverageTime)
  @OutputTimeUnit(TimeUnit.MICROSECONDS)
  public int splitAndTransferIntVector() {
    IntVector toVector = new IntVector("intVector", allocator);
    toVector.setValueCount(VECTOR_LENGTH);
    TransferPair transferPair = intVector.makeTransferPair(toVector);
    transferPair.splitAndTransfer(0, VECTOR_LENGTH);
    toVector.close();
    return 0;
  }

  @Benchmark
  @BenchmarkMode(Mode.AverageTime)
  @OutputTimeUnit(TimeUnit.MICROSECONDS)
  public int splitAndTransferVarcharVector() {
    VarCharVector toVector = new VarCharVector("varcharVector", allocator);
    toVector.setValueCount(VECTOR_LENGTH);
    TransferPair transferPair = varCharVector.makeTransferPair(toVector);
    transferPair.splitAndTransfer(0, VECTOR_LENGTH);
    toVector.close();
    return 0;
  }

  public static void main(String [] args) throws RunnerException {
    Options opt = new OptionsBuilder()
        .include(TransferPairBenchmarks.class.getSimpleName())
        .forks(1)
        .build();

    new Runner(opt).run();
  }
}
