// Licensed to the Apache Software Foundation (ASF) under one
// or more contributor license agreements.  See the NOTICE file
// distributed with this work for additional information
// regarding copyright ownership.  The ASF licenses this file
// to you under the Apache License, Version 2.0 (the
// "License"); you may not use this file except in compliance
// with the License.  You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing,
// software distributed under the License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
// KIND, either express or implied.  See the License for the
// specific language governing permissions and limitations
// under the License.

const { taskName } = require('./util');

const createTask = ((taskFn) => ((target, format, ...args) => {
  // Give the memoized fn a displayName so gulp's output is easier to follow.
  const fn = () => taskFn(target, format, ...args);
  fn.displayName = `${taskFn.name || ``}:${taskName(target, format, ...args)}:task`;
  return fn;
}));

const memoizeTask = ((cache, taskFn) => ((target, format, ...args) => {
    // Give the memoized fn a displayName so gulp's output is easier to follow.
    const fn = () => (
      cache[taskName(target, format)] || (
      cache[taskName(target, format)] = taskFn(target, format, ...args)));
    fn.displayName = `${taskFn.name || ``}:${taskName(target, format, ...args)}:task`;
    return fn;
}));

module.exports = memoizeTask;
module.exports.createTask = createTask;
module.exports.memoizeTask = memoizeTask;
